@@ -20,9 +20,17 @@
  * Boston, MA 02110-1301, USA.
  */
 
+#if HAS_STD_FILESYSTEM
+#if HAS_STD_FILESYSTEM_EXPERIMENTAL
+#include <experimental/filesystem>
+#else
+#include <filesystem>
+#endif
+#else
 #include <boost/filesystem/operations.hpp>   // for create_directories, exists
 #include <boost/filesystem/path.hpp>         // for path, operator<<
 #include <boost/filesystem/path_traits.hpp>  // for filesystem
+#endif
 #include <stddef.h>                          // for size_t
 #include <sys/stat.h>                        // for stat
 #include <volk/volk_prefs.h>                 // for volk_get_config_path
@@ -38,8 +46,15 @@
 #include "volk_option_helpers.h"             // for option_list, option_t
 #include "volk_profile.h"
 
-
+#if HAS_STD_FILESYSTEM
+#if HAS_STD_FILESYSTEM_EXPERIMENTAL
+namespace fs = std::experimental::filesystem;
+#else
+namespace fs = std::filesystem;
+#endif
+#else
 namespace fs = boost::filesystem;
+#endif
 
 volk_test_params_t test_params(1e-6f, 327.f, 131071, 1987, false, "");
 
@@ -70,7 +85,7 @@ int main(int argc, char *argv[]) {
     profile_options.add((option_t("json", "j", "Write results to JSON file named as argument value", set_json)));
     profile_options.add((option_t("path", "p", "Specify the volk_config path", set_volk_config)));
     profile_options.parse(argc, argv);
-    
+
     if (profile_options.present("help")) {
         return 0;
     }
@@ -151,12 +166,17 @@ int main(int argc, char *argv[]) {
     else {
         std::cout << "Warning: this was a dry-run. Config not generated" << std::endl;
     }
+    return 0;
 }
 
 void read_results(std::vector<volk_test_results_t> *results)
 {
     char path[1024];
-    volk_get_config_path(path);
+    volk_get_config_path(path, true);
+    if(path[0] == 0){
+        std::cout << "No prior test results found ..." << std::endl;
+        return;
+    }
 
     read_results(results, std::string(path));
 }
@@ -177,9 +197,8 @@ void read_results(std::vector<volk_test_results_t> *results, std::string path)
             std::vector<std::string> single_kernel_result;
             std::string config_str(config_line);
             std::size_t str_size = config_str.size();
-            std::size_t found = 1;
-
-            found = config_str.find(' ');
+            std::size_t found = config_str.find(' ');
+            
             // Split line by spaces
             while(found && found < str_size) {
                 found = config_str.find(' ');
@@ -213,7 +232,11 @@ void read_results(std::vector<volk_test_results_t> *results, std::string path)
 void write_results(const std::vector<volk_test_results_t> *results, bool update_result)
 {
     char path[1024];
-    volk_get_config_path(path);
+    volk_get_config_path(path, false);
+    if(path[0] == 0){
+        std::cout << "Aborting 'No config save path found' ..." << std::endl;
+        return;
+    }
 
     write_results( results, update_result, std::string(path));
 }
@@ -227,10 +250,10 @@ void write_results(const std::vector<volk_test_results_t> *results, bool update_
      * These
      */
     const fs::path config_path(path);
-    if (! fs::exists(config_path.branch_path()))
+    if (! fs::exists(config_path.parent_path()))
     {
-        std::cout << "Creating " << config_path.branch_path() << "..." << std::endl;
-        fs::create_directories(config_path.branch_path());
+        std::cout << "Creating " << config_path.parent_path() << "..." << std::endl;
+        fs::create_directories(config_path.parent_path());
     }
 
     std::ofstream config;
@@ -308,5 +331,3 @@ void write_json(std::ofstream &json_file, std::vector<volk_test_results_t> resul
     json_file << " ]" << std::endl;
     json_file << "}" << std::endl;
 }
-
-
