@@ -8,6 +8,7 @@
 #include <stdint.h>                                 // for uint16_t, uint64_t
 #include <sys/time.h>                               // for CLOCKS_PER_SEC
 #include <sys/types.h>                              // for int16_t, int32_t
+#include <chrono>
 #include <cmath>                                    // for sqrt, fabs, abs
 #include <cstring>                                  // for memcpy, memset
 #include <ctime>                                    // for clock
@@ -15,29 +16,35 @@
 #include <iostream>                                 // for cout, cerr
 #include <limits>                                   // for numeric_limits
 #include <map>                                      // for map, map<>::mappe...
+#include <random>
 #include <vector>                                   // for vector, _Bit_refe...
 
-float uniform() {
-  return 2.0f * ((float) rand() / RAND_MAX - 0.5f);	// uniformly (-1, 1)
-}
-
-template <class t>
-void random_floats (t *buf, unsigned n)
+template <typename T>
+void random_floats(void *buf, unsigned int n, std::default_random_engine& rnd_engine)
 {
-  for (unsigned i = 0; i < n; i++)
-    buf[i] = uniform ();
+    T *array = static_cast<T*>(buf);
+    std::uniform_real_distribution<T> uniform_dist(T(-1), T(1));
+    for(unsigned int i = 0; i < n; i++) {
+        array[i] = uniform_dist(rnd_engine);
+    }
 }
 
 void load_random_data(void *data, volk_type_t type, unsigned int n) {
+    std::random_device rnd_device;
+    std::default_random_engine rnd_engine(rnd_device());
     if(type.is_complex) n *= 2;
     if(type.is_float) {
-        if(type.size == 8) random_floats<double>((double *)data, n);
-        else random_floats<float>((float *)data, n);
+        if(type.size == 8) {
+            random_floats<double>(data, n, rnd_engine);
+        } else {
+            random_floats<float> (data, n, rnd_engine);
+        }
     } else {
         float int_max = float(uint64_t(2) << (type.size*8));
         if(type.is_signed) int_max /= 2.0;
+        std::uniform_real_distribution<float> uniform_dist(-int_max, int_max);
         for(unsigned int i=0; i<n; i++) {
-            float scaled_rand = (((float) (rand() - (RAND_MAX/2))) / static_cast<float>((RAND_MAX/2))) * int_max;
+            float scaled_rand = uniform_dist(rnd_engine);
             //man i really don't know how to do this in a more clever way, you have to cast down at some point
             switch(type.size) {
             case 8:
@@ -115,7 +122,7 @@ volk_type_t volk_type_from_string(std::string name) {
     }
     //will throw if malformed
     int size = volk_lexical_cast<int>(name.substr(0, last_size_pos+1));
-    
+
     assert(((size % 8) == 0) && (size <= 64) && (size != 0));
     type.size = size/8; //in bytes
 
@@ -134,7 +141,7 @@ volk_type_t volk_type_from_string(std::string name) {
             type.is_signed = false;
             break;
         default:
-            throw;
+            throw std::string("Error: no such type: '") + name[i] + "'";
         }
     }
 
@@ -182,7 +189,7 @@ static void get_signatures_from_name(std::vector<volk_type_t> &inputsig,
             if(side == SIDE_INPUT) inputsig.push_back(type);
             else outputsig.push_back(type);
         } catch (...){
-            if(token[0] == 'x' && (token.size() > 1) && (token[1] > '0' || token[1] < '9')) { //it's a multiplier
+            if(token[0] == 'x' && (token.size() > 1) && (token[1] > '0' && token[1] < '9')) { //it's a multiplier
                 if(side == SIDE_INPUT) assert(inputsig.size() > 0);
                 else assert(outputsig.size() > 0);
                 int multiplier = volk_lexical_cast<int>(token.substr(1, token.size()-1)); //will throw if invalid
@@ -247,28 +254,38 @@ inline void run_cast_test3_s32fc(volk_fn_3arg_s32fc func, std::vector<void *> &b
 }
 
 template <class t>
-bool fcompare(t *in1, t *in2, unsigned int vlen, float tol) {
+bool fcompare(t *in1, t *in2, unsigned int vlen, float tol, bool absolute_mode) {
     bool fail = false;
     int print_max_errs = 10;
     for(unsigned int i=0; i<vlen; i++) {
-        // for very small numbers we'll see round off errors due to limited
-        // precision. So a special test case...
-        if(fabs(((t *)(in1))[i]) < 1e-30) {
-            if( fabs( ((t *)(in2))[i] ) > tol )
-            {
+        if (absolute_mode) {
+            if (fabs(((t *)(in1))[i] - ((t *)(in2))[i]) > tol) {
                 fail=true;
                 if(print_max_errs-- > 0) {
                     std::cout << "offset " << i << " in1: " << t(((t *)(in1))[i]) << " in2: " << t(((t *)(in2))[i]);
                     std::cout << " tolerance was: " << tol << std::endl;
                 }
             }
-        }
-        // the primary test is the percent different greater than given tol
-        else if(fabs(((t *)(in1))[i] - ((t *)(in2))[i])/fabs(((t *)in1)[i]) > tol) {
-            fail=true;
-            if(print_max_errs-- > 0) {
-                std::cout << "offset " << i << " in1: " << t(((t *)(in1))[i]) << " in2: " << t(((t *)(in2))[i]);
-                std::cout << " tolerance was: " << tol << std::endl;
+        } else {
+            // for very small numbers we'll see round off errors due to limited
+            // precision. So a special test case...
+            if(fabs(((t *)(in1))[i]) < 1e-30) {
+                if( fabs( ((t *)(in2))[i] ) > tol )
+                {
+                    fail=true;
+                    if(print_max_errs-- > 0) {
+                    std::cout << "offset " << i << " in1: " << t(((t *)(in1))[i]) << " in2: " << t(((t *)(in2))[i]);
+                        std::cout << " tolerance was: " << tol << std::endl;
+                    }
+                }
+            }
+            // the primary test is the percent different greater than given tol
+            else if(fabs(((t *)(in1))[i] - ((t *)(in2))[i])/fabs(((t *)in1)[i]) > tol) {
+                fail=true;
+                if(print_max_errs-- > 0) {
+                    std::cout << "offset " << i << " in1: " << t(((t *)(in1))[i]) << " in2: " << t(((t *)(in2))[i]);
+                    std::cout << " tolerance was: " << tol << std::endl;
+                }
             }
         }
     }
@@ -277,10 +294,22 @@ bool fcompare(t *in1, t *in2, unsigned int vlen, float tol) {
 }
 
 template <class t>
-bool ccompare(t *in1, t *in2, unsigned int vlen, float tol) {
+bool ccompare(t *in1, t *in2, unsigned int vlen, float tol, bool absolute_mode) {
+    if (absolute_mode) {
+      std::cout << "ccompare does not support absolute mode" << std::endl;
+      return true;
+    }
     bool fail = false;
     int print_max_errs = 10;
     for(unsigned int i=0; i<2*vlen; i+=2) {
+        if (std::isnan(in1[i]) || std::isnan(in1[i+1]) || std::isnan(in2[i]) || std::isnan(in2[i+1])
+                || std::isinf(in1[i]) || std::isinf(in1[i+1]) || std::isinf(in2[i]) || std::isinf(in2[i+1])) {
+            fail=true;
+            if(print_max_errs-- > 0) {
+                std::cout << "offset " << i/2 << " in1: " << in1[i] << " + " << in1[i+1] << "j  in2: " << in2[i] << " + " << in2[i+1] << "j";
+                std::cout << " tolerance was: " << tol << std::endl;
+            }
+        }
         t diff[2] = { in1[i] - in2[i], in1[i+1] - in2[i+1] };
         t err  = std::sqrt(diff[0] * diff[0] + diff[1] * diff[1]);
         t norm = std::sqrt(in1[i] * in1[i] + in1[i+1] * in1[i+1]);
@@ -311,7 +340,11 @@ bool ccompare(t *in1, t *in2, unsigned int vlen, float tol) {
 }
 
 template <class t>
-bool icompare(t *in1, t *in2, unsigned int vlen, unsigned int tol) {
+bool icompare(t *in1, t *in2, unsigned int vlen, unsigned int tol, bool absolute_mode) {
+    if (absolute_mode) {
+      std::cout << "icompare does not support absolute mode" << std::endl;
+      return true;
+    }
     bool fail = false;
     int print_max_errs = 10;
     for(unsigned int i=0; i<vlen; i++) {
@@ -354,7 +387,7 @@ bool run_volk_tests(volk_func_desc_t desc,
 {
     return run_volk_tests(desc, manual_func, name, test_params.tol(), test_params.scalar(),
         test_params.vlen(), test_params.iter(), results, puppet_master_name,
-        test_params.benchmark_mode());
+        test_params.absolute_mode(), test_params.benchmark_mode());
 }
 
 bool run_volk_tests(volk_func_desc_t desc,
@@ -366,6 +399,7 @@ bool run_volk_tests(volk_func_desc_t desc,
                     unsigned int iter,
                     std::vector<volk_test_results_t> *results,
                     std::string puppet_master_name,
+                    bool absolute_mode,
                     bool benchmark_mode
 ) {
     // Initialize this entry in results vector
@@ -446,10 +480,10 @@ bool run_volk_tests(volk_func_desc_t desc,
 
     //now run the test
     vlen = vlen - vlen_twiddle;
-    clock_t start, end;
+    std::chrono::time_point<std::chrono::system_clock> start, end;
     std::vector<double> profile_times;
     for(size_t i = 0; i < arch_list.size(); i++) {
-        start = clock();
+        start = std::chrono::system_clock::now();
 
         switch(both_sigs.size()) {
             case 1:
@@ -493,9 +527,10 @@ bool run_volk_tests(volk_func_desc_t desc,
                 break;
         }
 
-        end = clock();
-        double arch_time = 1000.0 * (double)(end-start)/(double)CLOCKS_PER_SEC;
-        std::cout << arch_list[i] << " completed in " << arch_time << "ms" << std::endl;
+        end = std::chrono::system_clock::now();
+        std::chrono::duration<double> elapsed_seconds = end - start;
+        double arch_time = 1000.0 * elapsed_seconds.count();
+        std::cout << arch_list[i] << " completed in " << arch_time << " ms" << std::endl;
         volk_test_time_t result;
         result.name = arch_list[i];
         result.time = arch_time;
@@ -527,15 +562,15 @@ bool run_volk_tests(volk_func_desc_t desc,
                 if(both_sigs[j].is_float) {
                     if(both_sigs[j].size == 8) {
                         if (both_sigs[j].is_complex) {
-                            fail = ccompare((double *) test_data[generic_offset][j], (double *) test_data[i][j], vlen, tol_f);
+                            fail = ccompare((double *) test_data[generic_offset][j], (double *) test_data[i][j], vlen, tol_f, absolute_mode);
                         } else {
-                            fail = fcompare((double *) test_data[generic_offset][j], (double *) test_data[i][j], vlen, tol_f);
+                            fail = fcompare((double *) test_data[generic_offset][j], (double *) test_data[i][j], vlen, tol_f, absolute_mode);
                         }
                     } else {
                         if (both_sigs[j].is_complex) {
-                            fail = ccompare((float *) test_data[generic_offset][j], (float *) test_data[i][j], vlen, tol_f);
+                            fail = ccompare((float *) test_data[generic_offset][j], (float *) test_data[i][j], vlen, tol_f, absolute_mode);
                         } else {
-                            fail = fcompare((float *) test_data[generic_offset][j], (float *) test_data[i][j], vlen, tol_f);
+                            fail = fcompare((float *) test_data[generic_offset][j], (float *) test_data[i][j], vlen, tol_f, absolute_mode);
                         }
                     }
                 } else {
@@ -543,41 +578,41 @@ bool run_volk_tests(volk_func_desc_t desc,
                     switch(both_sigs[j].size) {
                     case 8:
                         if(both_sigs[j].is_signed) {
-                            fail = icompare((int64_t *) test_data[generic_offset][j], (int64_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((int64_t *) test_data[generic_offset][j], (int64_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         } else {
-                            fail = icompare((uint64_t *) test_data[generic_offset][j], (uint64_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((uint64_t *) test_data[generic_offset][j], (uint64_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         }
                         break;
                     case 4:
                         if(both_sigs[j].is_complex) {
                             if(both_sigs[j].is_signed) {
-                                fail = icompare((int16_t *) test_data[generic_offset][j], (int16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                                fail = icompare((int16_t *) test_data[generic_offset][j], (int16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                             } else {
-                                fail = icompare((uint16_t *) test_data[generic_offset][j], (uint16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                                fail = icompare((uint16_t *) test_data[generic_offset][j], (uint16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                             }
                         }
                         else {
                             if (both_sigs[j].is_signed) {
                                 fail = icompare((int32_t *) test_data[generic_offset][j], (int32_t *) test_data[i][j],
-                                                vlen * (both_sigs[j].is_complex ? 2 : 1), tol_i);
+                                                vlen * (both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                             } else {
                                 fail = icompare((uint32_t *) test_data[generic_offset][j], (uint32_t *) test_data[i][j],
-                                                vlen * (both_sigs[j].is_complex ? 2 : 1), tol_i);
+                                                vlen * (both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                             }
                         }
                         break;
                     case 2:
                         if(both_sigs[j].is_signed) {
-                            fail = icompare((int16_t *) test_data[generic_offset][j], (int16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((int16_t *) test_data[generic_offset][j], (int16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         } else {
-                            fail = icompare((uint16_t *) test_data[generic_offset][j], (uint16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((uint16_t *) test_data[generic_offset][j], (uint16_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         }
                         break;
                     case 1:
                         if(both_sigs[j].is_signed) {
-                            fail = icompare((int8_t *) test_data[generic_offset][j], (int8_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((int8_t *) test_data[generic_offset][j], (int8_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         } else {
-                            fail = icompare((uint8_t *) test_data[generic_offset][j], (uint8_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i);
+                            fail = icompare((uint8_t *) test_data[generic_offset][j], (uint8_t *) test_data[i][j], vlen*(both_sigs[j].is_complex ? 2 : 1), tol_i, absolute_mode);
                         }
                         break;
                     default:
@@ -586,7 +621,7 @@ bool run_volk_tests(volk_func_desc_t desc,
                 }
                 if(fail) {
                     volk_test_time_t *result = &results->back().results[arch_list[i]];
-                    result->pass = !fail;
+                    result->pass = false;
                     fail_global = true;
                     std::cout << name << ": fail on arch " << arch_list[i] << std::endl;
                 }
